module Data.IP.Internal ( IPv4(..)
                        , IPv6(..)
                        , AddrRange(..)
                        , IPv4Addr
                        , IPv6Addr
                        ) where

import Data.IP.Addr
import Data.IP.Range
